import asyncio
from collections import Callable
from inspect import iscoroutinefunction
from typing import Any


async def run_smart_async(func: Callable[..., Any], *args: Any) -> Any:
    if iscoroutinefunction(func):
        returned = await func(*args)
    else:
        loop = asyncio.get_event_loop()
        returned = await loop.run_in_executor(None, func, *args)
    return returned



from enum import Enum
from pathlib import Path
from typing import Awaitable, Callable, Optional, Union

from databases import Database
from loguru import logger
from pydantic import BaseConfig, BaseModel
from sqlalchemy import MetaData

from aopi_index_builder.exceptions import UserHasNoPermissions, UserWasNotFound


class AopiContextBase(BaseModel):
    """
    :database: context to use databases.
    :metadata: sqlalchemy metadata for table creation.
    :main_dir: directory to store packages.
    """

    database: Database
    metadata: MetaData
    main_dir: Path
    enable_users: bool
    get_user_id_function: Callable[["AopiContextBase", str, str], Awaitable[int]]
    check_user_permission: Callable[
        ["AopiContextBase", str, Optional[int], Union[str, Enum]], Awaitable[bool]
    ]

    class Config(BaseConfig):
        arbitrary_types_allowed = True
        allow_mutation = False


class PackageContext(BaseModel):
    """
    :prefix: base url prefix
    :packages_dir: your own directory to store packages.
    """

    prefix: str
    package_name: str
    packages_dir: Path

    class Config(BaseConfig):
        allow_mutation = False


class AopiContext(AopiContextBase, PackageContext):
    async def get_user_id(self, username: str, password: str) -> Optional[int]:
        """
        Get user id by username and password.

        This function will look up user by credentials in aopi database.
        If user system is disabled by aopi, then None would be returned.
        If user was not found ..py:exception::`UserWasNotFound`
        exception would be raised.

        :param username: user's username
        :param password: user's password
        :return: id, None or exception
        :exception: UserWasNotFound
        """
        if not self.enable_users:
            return None
        user_id = None
        try:
            user_id = await self.get_user_id_function(username, password)
        except Exception as e:
            logger.exception(e)
        if user_id is None:
            raise UserWasNotFound()
        return user_id

    async def has_permission(
        self, user_id: Optional[int], role: Union[str, Enum]
    ) -> None:
        """
        Check if user has permissions.
        You can pass either enum or name of the role to check.

        Usage is following:

            >>> from typing import Any
            >>> async def your_function() -> Any:
            >>>     context = get_context()
            >>>     uid = await context.get_user_id("username", "password")
            >>>     await context.has_permission(uid, "role")
            >>>     ... # do something, you're safe

        The key thing is that if user has no permission or either cannot be found
        the exception will stop your function from execution.

        :param user_id: user's unique id from aopi
        :param role: required role for action
        :return: bool or raise an exception
        :exception: UserHasNoPermissions
        """
        if not self.enable_users:
            return
        if user_id is None:
            raise UserHasNoPermissions()
        try:
            if not await self.check_user_permission(
                self.package_name,
                user_id,
                role.value if isinstance(role, Enum) else role,
            ):
                raise UserHasNoPermissions()
            return
        except Exception as e:
            logger.exception(e)
        except UserHasNoPermissions:
            pass
        raise UserHasNoPermissions()

    class Config(BaseConfig):
        arbitrary_types_allowed = True


__base_ctx: Optional[AopiContextBase] = None
__package_ctx: Optional[PackageContext] = None


def init_context(base: AopiContextBase) -> None:
    """
    Initialize base context for plugins.

    This context parameters are the same for all plugins.
    And this context must be initialized manually in aopi application.

    :param base: this variable contains database and metadata for creating tables.
    """
    global __base_ctx
    __base_ctx = base


def get_base_ctx() -> AopiContextBase:
    """
    Just return the base context for all plugins.
    :return: context
    """
    global __base_ctx
    if __base_ctx is None:
        raise ValueError("Base context is not initialized.")
    return __base_ctx


def init_package_ctx(ctx: PackageContext) -> None:
    """
    This context is created for individual plugin.

    :param ctx: new package context.
    """
    global __package_ctx
    __package_ctx = ctx


def get_context() -> AopiContext:
    """
    Just returns the context for new plugin.

    It has base prefix for this plugin and
    packages dir specifically for current plugin. So you can store anything in it
    without breaking other plugins.

    You can call this function many times and your context will be the same.
    Also, you can't override your context. It's immutable for other plugins safety.

    :return: current plugin context.
    """
    base_ctx = get_base_ctx()
    global __package_ctx
    if __package_ctx is None:
        raise ValueError("Context is not initialized.")

    return AopiContext(**base_ctx.dict(), **__package_ctx.dict())



import io
import os
from contextlib import redirect_stdout
from enum import Enum
from typing import Any, Awaitable, Callable, List, Optional, Type, Union

import entrypoints
from fastapi import APIRouter
from loguru import logger
from pydantic import BaseConfig, BaseModel

from aopi_index_builder.context import PackageContext, get_base_ctx, init_package_ctx
from aopi_index_builder.schema import FullPackageInfo, PackagePreview, PackageVersion


class PackageIndex(BaseModel):
    """
    The main class for creating new package indices.
    """

    router: APIRouter
    target_language: str
    db_models: List[Type[Any]] = []
    roles: Type[Enum]
    plugin_help: Optional[str] = None
    find_packages_func: Callable[
        [Optional[int], str, int],
        Union[Awaitable[List[PackagePreview]], List[PackagePreview]],
    ]
    get_package_info_func: Callable[
        [Optional[int], Any], Union[Awaitable[FullPackageInfo], List[FullPackageInfo]]
    ]
    get_versions_func: Callable[
        [Optional[int], Any],
        Union[Awaitable[List[PackageVersion]], List[PackageVersion]],
    ]

    def __init__(
        self,
        router: APIRouter,
        target_language: str,
        roles: Type[Enum],
        find_packages_func: Callable[
            [Optional[int], str, int],
            Union[Awaitable[List[PackagePreview]], List[PackagePreview]],
        ],
        get_package_info_func: Callable[
            [Optional[int], Any],
            Union[Awaitable[FullPackageInfo], List[FullPackageInfo]],
        ],
        get_versions_func: Callable[
            [Optional[int], Any],
            Union[Awaitable[List[PackageVersion]], List[PackageVersion]],
        ],
        db_models: Optional[List[Type[Any]]] = None,
        plugin_help: Optional[str] = None,
    ):
        """
        Package index constructor.
        You must return instance of this class when creating new package index.

        :param router: router for package index api.
            All this routes appending to aopi routes to be used for integration with
            package managers.
        :param target_language: Name of target language or technology.
        :param roles: Roles for aopi rbac system to be used in this plugin.
        :param find_packages_func: function to find package in index.
            This function takes 4 positional arguments:
                * "current_user_id": Optional[int]
                * "package_name": str
                * "limit": int
                * "offset": int
        :param get_package_info_func: function to get full information about package.
            This function takes two arguments:
                * "current_user_id": Optional[int]
                * "package_id": Any.
        :param get_versions_func: function to get all versions of specific package.
            it takes two positional arguments:
                * "current_user_id": Optional[int]
                * "package_id": Any.
        :param db_models: used in this plugin.
            This thing is made to be sure, that all models for every plugin are created.
        :param plugin_help: This string is used as help
            for user about how to use provided api.
            For example:
            "To use aopi_python install packages as the following:
            pip install --index-url http://{aopi_url}/python/simple"

        """
        super(PackageIndex, self).__init__(
            router=router,
            target_language=target_language,
            roles=roles,
            db_models=db_models or [],
            plugin_help=plugin_help,
            find_packages_func=find_packages_func,
            get_package_info_func=get_package_info_func,
            get_versions_func=get_versions_func,
        )

    class Config(BaseConfig):
        arbitrary_types_allowed = True


class PluginInfo(BaseModel):
    prefix: str
    plugin_name: str
    roles: List[str]
    package_name: str
    package_version: str
    package_index: PackageIndex


def load_plugins() -> List[PluginInfo]:
    """
    Discover and load all plugins available in
    current environment.

    """
    indices = []
    base_ctx = get_base_ctx()
    for entrypoint in entrypoints.get_group_all("aopi_index"):
        plugin_name = entrypoint.name
        plugin_distro = entrypoint.distro
        plugin_package_dir = base_ctx.main_dir.joinpath(plugin_name)
        if not plugin_package_dir.exists():
            os.makedirs(plugin_package_dir)
        plugin_prefix = f"/{plugin_name}"
        init_package_ctx(
            PackageContext(
                prefix=plugin_prefix,
                package_name=plugin_distro.name,
                packages_dir=plugin_package_dir,
            )
        )
        logger.debug(f"Loading {plugin_name}")
        buffer = io.StringIO()
        with redirect_stdout(buffer):
            try:
                index_factory = entrypoint.load()
                index = index_factory()
                if not isinstance(index, PackageIndex):
                    logger.error("Plugin has returned wrong type.")
                    logger.debug(f"Expected: PackageIndex. Actual: {index.__class__}")
                    continue
                indices.append(
                    PluginInfo(
                        prefix=plugin_prefix,
                        plugin_name=plugin_name,
                        roles=[role.value for role in index.roles],
                        package_name=plugin_distro.name,
                        package_version=plugin_distro.version,
                        package_index=index,
                    )
                )
            except Exception as e:
                logger.error(f"Can't load plugin {plugin_name}")
                logger.exception(e)
            logger.debug(f"{plugin_name} captured output: \n{buffer.getvalue()}")
    return indices



from typing import Any, Dict, Optional

from pydantic import BaseModel


class PluginInfo(BaseModel):
    plugin_name: str
    language: str


class PackagePreview(BaseModel):
    id: Any
    name: str
    short_description: Optional[str]


class PluginPackagePreview(PluginInfo, PackagePreview):
    ...


class FullPackageInfo(PackagePreview):
    description: Optional[str]
    metadata: Dict[str, Any]
    last_version: str


class PluginFullPackageInfo(PluginInfo, FullPackageInfo):
    ...


class PackageVersion(BaseModel):
    version: str
    yanked: bool
    metadata: Dict[str, Any]



from fastapi import HTTPException


class UserError(HTTPException):
    def __init__(self) -> None:
        self.status_code = 401
        self.detail = "Something bad happened with user"


class UserWasNotFound(UserError):
    def __init__(self) -> None:
        super(UserWasNotFound, self).__init__()
        self.detail = "Wrong credentials"


class UserHasNoPermissions(UserError):
    def __init__(self) -> None:
        super(UserHasNoPermissions, self).__init__()
        self.detail = "You has no permission to do this"
        self.status_code = 403



from typing import Any, Dict, List, Optional

from fastapi import FastAPI
from loguru import logger
from pydantic import BaseModel

from aopi_index_builder import AopiContextBase, PluginInfo, init_context, load_plugins
from aopi_index_builder.exceptions import UserError
from aopi_index_builder.schema import (
    PackageVersion,
    PluginFullPackageInfo,
    PluginPackagePreview,
)
from aopi_index_builder.utils import run_smart_async


class PluginRole(BaseModel):
    plugin_name: str
    role: str


class PluginManager:
    def __init__(self, context: AopiContextBase) -> None:
        self.context = context
        self.plugins_map: Dict[str, PluginInfo] = dict()

    def load(self) -> None:
        init_context(self.context)
        plugins = load_plugins()
        self.plugins_map = {plugin.package_name: plugin for plugin in plugins}

    def get_roles(self) -> List[PluginRole]:
        roles: List[PluginRole] = list()
        for plugin in self.plugins_map.values():
            roles.extend(
                map(
                    lambda x: PluginRole(plugin_name=plugin.package_name, role=x),
                    plugin.roles,
                )
            )
        return roles

    def get_languages(self) -> List[str]:
        return list(
            map(lambda x: x.package_index.target_language, self.plugins_map.values())
        )

    async def find_package(
        self,
        *,
        user_id: Optional[int],
        package_name: str,
        limit: int,
        page: int,
        language: Optional[str] = None,
    ) -> List[PluginPackagePreview]:
        plugins_count = len(self.plugins_map.values())
        packages: List[PluginPackagePreview] = list()
        plugin_limit = limit // plugins_count
        offset = plugin_limit * page
        for plugin in self.plugins_map.values():
            if (
                language is not None
                and plugin.package_index.target_language.lower() != language.lower()
            ):
                continue
            func = plugin.package_index.__dict__["find_packages_func"]
            try:
                plugin_packages = await run_smart_async(
                    func, user_id, package_name, plugin_limit, offset
                )
                packages.extend(
                    map(
                        lambda package: PluginPackagePreview(
                            **package.dict(),
                            plugin_name=plugin.package_name,
                            language=plugin.package_index.target_language,
                        ),
                        plugin_packages,
                    )
                )
            except UserError as e:
                logger.debug(e)
                continue
        return packages

    async def get_package_versions(
        self, *, user_id: Optional[int], plugin_name: str, package_id: Any
    ) -> List[PackageVersion]:
        plugin = self.plugins_map.get(plugin_name)
        if not plugin:
            return list()
        func = plugin.package_index.__dict__["get_versions_func"]
        try:
            return await run_smart_async(func, user_id, package_id)
        except UserError as e:
            logger.debug(e)
        return list()

    async def get_package_info(
        self, *, user_id: str, plugin_name: str, package_id: Any
    ) -> Optional[PluginFullPackageInfo]:
        plugin = self.plugins_map.get(plugin_name)
        if not plugin:
            return None
        func = plugin.package_index.__dict__["get_package_info_func"]
        try:
            package_info = await run_smart_async(func, user_id, package_id)
            return PluginFullPackageInfo(
                **package_info.dict(),
                plugin_name=plugin.package_name,
                language=plugin.package_index.target_language,
            )
        except UserError as e:
            logger.debug(e)
        return None

    def add_routes(self, app: FastAPI) -> None:
        for plugin in self.plugins_map.values():
            index = plugin.package_index
            logger.debug(
                f"Enabling plugin {plugin.plugin_name} "
                f"from {plugin.package_name}:{plugin.package_version}"
            )
            app.include_router(
                router=index.router, prefix=plugin.prefix, tags=[plugin.package_name]
            )



# -*- coding: utf-8 -*-
from setuptools import setup

packages = \
['aopi_index_builder']

package_data = \
{'': ['*']}

install_requires = \
['Jinja2>=2.11.2,<3.0.0',
 'SQLAlchemy>=1.3.22,<2.0.0',
 'databases>=0.4.1,<0.5.0',
 'entrypoints>=0.3,<0.4',
 'fastapi>=0.63.0,<0.64.0',
 'loguru>=0.5.3,<0.6.0',
 'pydantic>=1.7.3,<2.0.0']

setup_kwargs = {
    'name': 'aopi-index-builder',
    'version': '0.1.40',
    'description': 'Package to build new package indices for AOPI',
    'long_description': None,
    'author': 'Pavel Kirilin',
    'author_email': 'win10@list.ru',
    'maintainer': 'Pavel Kirilin',
    'maintainer_email': 'win10@list.ru',
    'url': None,
    'packages': packages,
    'package_data': package_data,
    'install_requires': install_requires,
    'python_requires': '>=3.9,<4.0',
}


setup(**setup_kwargs)



from aopi_index_builder.context import (
    AopiContext,
    AopiContextBase,
    get_context,
    init_context,
)
from aopi_index_builder.package_index import PackageIndex, PluginInfo, load_plugins
from aopi_index_builder.plugin_manager import PluginManager
from aopi_index_builder.schema import (
    FullPackageInfo,
    PackagePreview,
    PackageVersion,
    PluginFullPackageInfo,
    PluginPackagePreview,
)

__all__ = [
    "AopiContextBase",
    "AopiContext",
    "get_context",
    "init_context",
    "PackageIndex",
    "PluginInfo",
    "load_plugins",
    "PluginManager",
    "PackagePreview",
    "FullPackageInfo",
    "PackageVersion",
    "PluginPackagePreview",
    "PluginFullPackageInfo",
]
